# -*- coding: utf-8 -*-

import logging
import struct
import threading
from concurrent.futures import FIRST_EXCEPTION, ThreadPoolExecutor, as_completed, wait
from typing import Any, Dict, Iterable, List, Tuple

from click import progressbar
from evernote.edam.notestore.ttypes import SyncChunk
from evernote.edam.type.ttypes import LinkedNotebook, Note

from evernote_backup.cli_app_util import chunks, get_progress_output
from evernote_backup.evernote_client_sync import EvernoteClientSync
from evernote_backup.evernote_client_util import NotebookAuth
from evernote_backup.note_storage import NoteForSync, SqliteStorage

logger = logging.getLogger(__name__)


THREAD_CHUNK_SIZE = 1000


class WrongAuthUserError(Exception):
    """Raise when remote auth user is not the same as the one registered in database"""

    def __init__(self, local_user: str, remote_user: str) -> None:
        self.local_user = local_user
        self.remote_user = remote_user


class WorkerStopException(Exception):
    """Raise when workers are stopped"""


def get_note_size(note: Note) -> int:
    size = note.contentLength

    size += sum(r.data.size for r in note.resources or [])

    return int(size)


class NoteClientMemoryManager(object):
    def __init__(self, download_cache_memory_limit: int) -> None:
        self.memory_limit = download_cache_memory_limit * 1024 * 1024

        self.memory = 0
        self.memory_lock = threading.Lock()

        self.memory_cond = threading.Condition()

    def wait_till_enough_memory(self) -> None:
        with self.memory_cond:
            self.memory_cond.wait_for(self._is_enough_memory)

    def reset_memory(self) -> None:
        with self.memory_lock:
            self.memory = 0

        with self.memory_cond:
            self.memory_cond.notify_all()

    def add_note_size(self, note: Note) -> None:
        with self.memory_lock:
            self.memory += get_note_size(note)

    def sub_note_size(self, note: Note) -> None:
        with self.memory_lock:
            self.memory -= get_note_size(note)

        with self.memory_cond:
            if self._is_enough_memory():
                self.memory_cond.notify_all()

    def report_memory(self) -> None:
        with self.memory_lock:
            memory_percent = round(self.memory / self.memory_limit, 3)

        memory_total = self.memory_limit // (1024 * 1024)

        logger.debug(f"Memory consumed: {memory_percent}% [LIMIT {memory_total} MB]")

    def _is_enough_memory(self) -> bool:
        with self.memory_lock:
            return self.memory < self.memory_limit


class NoteClientWorker(object):
    def __init__(  # noqa: WPS211
        self,
        token: str,
        backend: str,
        network_error_retry_count: int,
        max_chunk_results: int,
        download_cache_memory_limit: int,
    ) -> None:
        self.stop = False
        self.token = token
        self.backend = backend
        self.network_error_retry_count = network_error_retry_count
        self.max_chunk_results = max_chunk_results

        self.memory_manager = NoteClientMemoryManager(download_cache_memory_limit)

        self._thread_data = threading.local()
        self._note_client: EvernoteClientSync

    def __call__(self, note_id: str, auth_data: NotebookAuth = None) -> Note:
        self.memory_manager.wait_till_enough_memory()

        if self.stop:
            raise WorkerStopException

        if auth_data is None:
            auth_data = NotebookAuth(token=self.token, shard="")

        client_id = auth_data.shard + auth_data.token

        try:
            self._note_client = self.clients[client_id]
        except KeyError:
            self._note_client = EvernoteClientSync(
                token=auth_data.token,
                backend=self.backend,
                network_error_retry_count=self.network_error_retry_count,
                max_chunk_results=self.max_chunk_results,
            )

            if auth_data.shard:
                self._note_client.shard = auth_data.shard
                self._note_client.shared_mode = True

            self.clients[client_id] = self._note_client

        note = self.download_note(note_id)

        self.memory_manager.add_note_size(note)
        self.memory_manager.report_memory()

        return note

    def download_note(self, note_id: str) -> Note:
        retry_count = 5

        for _ in range(retry_count):
            try:
                return self._note_client.get_note(note_id)
            except struct.error:
                logger.debug(
                    f"Remote server returned bad data"
                    f" while downloading note [{note_id}], retrying..."
                )

        raise RuntimeError(
            f"Failed to download note [{note_id}] after {retry_count} attempts!"
        )

    @property
    def clients(self) -> Any:
        try:
            return self._thread_data.note_clients
        except AttributeError:
            self._thread_data.note_clients = {}
            return self._thread_data.note_clients


class NoteSynchronizer(object):  # noqa: WPS214
    def __init__(
        self,
        note_client: EvernoteClientSync,
        note_storage: SqliteStorage,
        max_download_workers: int,
        download_cache_memory_limit: int,
    ) -> None:
        self._count_updated_notebooks = 0
        self._count_updated_notes = 0

        self._count_expunged_notebooks = 0
        self._count_expunged_linked_notebooks = 0
        self._count_expunged_notes = 0

        self.note_client = note_client
        self.storage = note_storage
        self.max_download_workers = max_download_workers

        self.note_worker = NoteClientWorker(
            self.note_client.token,
            self.note_client.backend,
            self.note_client.network_error_retry_count,
            self.note_client.max_chunk_results,
            download_cache_memory_limit,
        )
        self.linked_notebooks_auth: Dict[str, NotebookAuth] = {}

    def sync(self) -> None:
        self._raise_on_wrong_user()

        logger.info("Syncing user notebooks...")

        self._sync_chunks()

        if self.note_client.linked_notebooks:
            logger.info("Syncing linked notebooks...")

            for l_notebook in self.note_client.linked_notebooks.values():
                self._sync_linked_notebook(l_notebook)

        notes_to_sync = self.storage.notes.get_notes_for_sync()

        if notes_to_sync:
            logger.info("{0} note(s) to download...".format(len(notes_to_sync)))

            self._authorize_linked_notebooks_for_notes(notes_to_sync)
            self._download_scheduled_notes(notes_to_sync)

            self._count_updated_notes = len(notes_to_sync)

        report = [
            f"Updated or added notebooks: {self._count_updated_notebooks}",
            f"Updated or added notes: {self._count_updated_notes}",
            f"Expunged notebooks: {self._count_expunged_notebooks}",
            f"Expunged linked notebooks: {self._count_expunged_linked_notebooks}",
            f"Expunged notes: {self._count_expunged_notes}",
        ]

        for msg in report:
            logger.info(msg)

    def _authorize_linked_notebooks_for_notes(
        self, notes_to_sync: Tuple[NoteForSync, ...]
    ) -> None:
        linked_notebooks = {
            n.linked_notebook_guid for n in notes_to_sync if n.linked_notebook_guid
        }

        if linked_notebooks:
            logger.info(
                "Requesting access to {0} linked notebook(s)...".format(
                    len(linked_notebooks)
                )
            )

            for ln_guid in linked_notebooks:
                nb = self.storage.notebooks.get_notebook_by_linked_guid(ln_guid)
                auth_token = self.note_client.auth_linked_notebook(ln_guid, nb.guid)

                self.linked_notebooks_auth[ln_guid] = auth_token

    def _raise_on_wrong_user(self) -> None:
        remote_user = self.note_client.user
        local_user = self.storage.config.get_config_value("user")

        if remote_user != local_user:
            raise WrongAuthUserError(local_user, remote_user)

    def _sync_chunks(self) -> None:
        remote_usn = self.note_client.get_remote_usn()
        current_usn = int(self.storage.config.get_config_value("USN"))

        if remote_usn == current_usn:
            logger.info("User notebooks are up to date, nothing to sync!")
            return

        last_usn = current_usn
        with progressbar(
            length=remote_usn - current_usn,
            show_pos=True,
            file=get_progress_output(),
        ) as chunks_bar:
            for chunk in self.note_client.iter_sync_chunks(current_usn):
                self._process_chunk(chunk)

                self.storage.config.set_config_value("USN", str(chunk.chunkHighUSN))

                chunks_bar.update(chunk.chunkHighUSN - last_usn)
                last_usn = chunk.chunkHighUSN

    def _sync_linked_notebook(self, l_notebook: LinkedNotebook) -> None:
        current_usn = self.storage.notebooks.get_linked_notebook_usn(l_notebook.guid)

        l_notebook_chunks = self.note_client.iter_linked_notebook_sync_chunks(
            l_notebook, current_usn
        )

        for chunk in l_notebook_chunks:
            for notebook in chunk.notebooks or []:
                # Correct stack info is in LinkedNotebook
                notebook.stack = l_notebook.stack
                self.storage.notebooks.add_linked_notebook(l_notebook, notebook)

            self._process_chunk(chunk)

            self.storage.notebooks.set_linked_notebook_usn(
                l_notebook.guid, chunk.chunkHighUSN
            )

    def _process_chunk(self, chunk: SyncChunk) -> None:
        self._expunge(
            chunk.expungedNotebooks,
            chunk.expungedNotes,
            chunk.expungedLinkedNotebooks,
        )

        if chunk.notebooks:
            self.storage.notebooks.add_notebooks(chunk.notebooks)

            self._count_updated_notebooks += len(chunk.notebooks)

        if chunk.notes:
            self.storage.notes.add_notes_for_sync(chunk.notes)

    def _expunge(
        self,
        expunged_notebooks: List[str],
        expunged_notes: List[str],
        expunged_linked_notebooks: List[str],
    ) -> None:
        if expunged_notebooks:
            self.storage.notebooks.expunge_notebooks(expunged_notebooks)

            self._count_expunged_notebooks += len(expunged_notebooks)

        if expunged_notes:
            self.storage.notes.expunge_notes(expunged_notes)

            self._count_expunged_notes += len(expunged_notes)

        if expunged_linked_notebooks:
            for l_notebook_guid in expunged_linked_notebooks:
                self._expunge_linked_assoc(l_notebook_guid)

            self.storage.notebooks.expunge_linked_notebooks(expunged_linked_notebooks)

            self._count_expunged_linked_notebooks += len(expunged_linked_notebooks)

    def _expunge_linked_assoc(self, l_notebook_guid: str) -> None:
        try:
            notebook = self.storage.notebooks.get_notebook_by_linked_guid(
                l_notebook_guid
            )
        except ValueError:
            return

        self.storage.notebooks.expunge_notebooks((notebook.guid,))
        self.storage.notes.expunge_notes_by_notebook(notebook.guid)

    def _download_scheduled_notes(self, notes_to_sync: Tuple[NoteForSync, ...]) -> None:
        logger.info("Downloading {0} note(s)...".format(len(notes_to_sync)))
        logger.debug(f"Sync worker threads: {self.max_download_workers}")

        with ThreadPoolExecutor(max_workers=self.max_download_workers) as executor:
            with progressbar(
                length=len(notes_to_sync),
                show_pos=True,
                file=get_progress_output(),
                item_show_func=lambda x: x.title if x else "",  # type: ignore
            ) as notes_bar:
                for notes_to_sync_chunk in chunks(notes_to_sync, THREAD_CHUNK_SIZE):
                    self._process_download_chunk(
                        executor, notes_bar, notes_to_sync_chunk
                    )

    def _process_download_chunk(
        self,
        executor: Any,
        notes_bar: Any,
        notes_chunk: Iterable[NoteForSync],
    ) -> None:
        note_futures = {
            executor.submit(
                self.note_worker,
                n.guid,
                self.linked_notebooks_auth.get(n.linked_notebook_guid),  # type: ignore
            ): n.title
            for n in notes_chunk
        }

        try:
            for note_f in as_completed(note_futures):
                f_exc = note_f.exception()
                if f_exc is not None:
                    logger.critical(
                        f"Exception caught while downloading note"
                        f" '{note_futures[note_f]}'!"
                    )

                    raise f_exc

                note = note_f.result(timeout=120)  # noqa: WPS432
                self.storage.notes.add_note(note)

                self.note_worker.memory_manager.sub_note_size(note)

                notes_bar.update(1, note)

        except (KeyboardInterrupt, Exception):
            logger.warning("Aborting, please wait...")

            self.note_worker.stop = True
            self.note_worker.memory_manager.reset_memory()

            wait(note_futures, timeout=30, return_when=FIRST_EXCEPTION)  # noqa: WPS432

            raise



import functools
import inspect
import platform
from typing import Any, Optional, Union

from evernote.edam.error.ttypes import EDAMSystemException, EDAMUserException
from evernote.edam.notestore import NoteStore
from evernote.edam.userstore import UserStore
from thrift.protocol import TBinaryProtocol
from thrift.transport import THttpClient

from evernote_backup.evernote_client_classes import ClientV2
from evernote_backup.evernote_client_util import network_retry, raise_auth_error
from evernote_backup.token_util import get_token_shard


class EvernoteClientBase(object):
    def __init__(self, backend: str) -> None:
        self.backend = backend

        backends = {
            "evernote": "www.evernote.com",
            "evernote:sandbox": "sandbox.evernote.com",
            "china": "app.yinxiang.com",
            "china:sandbox": "sandbox.yinxiang.com",
        }

        self.service_host = backends[self.backend]

        if self.backend.startswith("china"):
            self.user_agent = "YXScript Windows/603139;"
        else:
            self.user_agent = "ENScript Windows/309091;"

        self.device_description = "evernote-backup"
        if platform.node():
            self.device_description += " [{0}]".format(platform.node())

    def _get_endpoint(self, path: str = "") -> str:
        return f"https://{self.service_host}/{path}"


class EvernoteClient(EvernoteClientBase):
    def __init__(
        self, backend: str, token: str, network_error_retry_count: int
    ) -> None:
        super().__init__(backend=backend)

        self.token = token
        self.shard = get_token_shard(token)

        self.network_error_retry_count = network_error_retry_count

        self._user: Optional[str] = None

    def verify_token(self) -> None:
        try:
            self.user_store.getUser()
        except (EDAMUserException, EDAMSystemException) as e:
            raise_auth_error(e)
            raise

    @property
    def user_store(self) -> "Store":
        user_store_uri = self._get_endpoint("edam/user")
        return Store(
            client_class=ClientV2,
            store_url=user_store_uri,
            token=self.token,
            user_agent=self.user_agent,
            network_error_retry_count=self.network_error_retry_count,
        )

    @property
    def note_store(self) -> "Store":
        return self.get_note_store()

    @property
    def user(self) -> str:
        if self._user is None:
            self._user = self.user_store.getUser().username
        return self._user

    def get_note_store(self, shard_id: str = None) -> "Store":
        if shard_id is None:
            shard_id = self.shard

        note_store_uri = self._get_endpoint(f"edam/note/{shard_id}")

        return Store(
            client_class=NoteStore.Client,
            store_url=note_store_uri,
            token=self.token,
            user_agent=self.user_agent,
            network_error_retry_count=self.network_error_retry_count,
        )


class Store(object):
    def __init__(
        self,
        client_class: Union[UserStore.Client, NoteStore.Client],
        store_url: str,
        user_agent: str,
        network_error_retry_count: int,
        token: Optional[str] = None,
    ):
        self.token = token
        self.user_agent = user_agent
        self.network_error_retry_count = network_error_retry_count

        self._client = self._get_thrift_client(client_class, store_url)

    def __getattr__(self, name: str) -> Any:
        target_method = getattr(self._client, name)

        if not callable(target_method):
            return target_method

        def wrapper(*args: Any, **kwargs: Any) -> Any:
            target_method_retry = network_retry(self.network_error_retry_count)(
                target_method
            )

            org_args = inspect.getfullargspec(target_method).args
            if len(org_args) == len(args) + 1:
                return target_method_retry(*args, **kwargs)
            elif self.token and "authenticationToken" in org_args:
                skip_args = ["self", "authenticationToken"]
                arg_names = [i for i in org_args if i not in skip_args]
                return functools.partial(
                    target_method_retry,
                    authenticationToken=self.token,
                )(
                    **dict(list(zip(arg_names, args))),
                )

            return target_method_retry(*args, **kwargs)

        return wrapper

    def _get_thrift_client(
        self,
        client_class: Union[UserStore.Client, NoteStore.Client],
        url: str,
    ) -> Union[UserStore.Client, NoteStore.Client]:
        http_client = THttpClient.THttpClient(url)
        http_client.setCustomHeaders(
            {
                "User-Agent": self.user_agent,
                "x-feature-version": 3,
                "accept": "application/x-thrift",
                "cache-control": "no-cache",
            }
        )

        thrift_protocol = TBinaryProtocol.TBinaryProtocol(http_client)
        return client_class(thrift_protocol)



import time
from http.client import HTTPException
from typing import Any, Callable, NamedTuple, Union

from evernote.edam.error.ttypes import (
    EDAMErrorCode,
    EDAMSystemException,
    EDAMUserException,
)


class NotebookAuth(NamedTuple):
    token: str
    shard: str


class EvernoteAuthError(Exception):
    """Evernote authentication error"""


def network_retry(network_error_retry_count: int) -> Callable:
    def decorator(fun: Callable) -> Callable:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            retry_count = network_error_retry_count

            for i in range(network_error_retry_count):
                try:
                    return fun(*args, **kwargs)
                except (HTTPException, ConnectionError):
                    if i == retry_count - 1:
                        raise
                    time.sleep(0.5)

        return wrapper

    return decorator


def raise_auth_error(exception: Union[EDAMSystemException, EDAMUserException]) -> None:
    messages = {
        EDAMErrorCode.BAD_DATA_FORMAT: {"authenticationToken": "Wrong token format!"},
        EDAMErrorCode.INVALID_AUTH: {
            "username": "Username not found!",
            "password": "Invalid password!",
            "oneTimeCode": "Invalid one-time code!",
            "authenticationToken": "Invalid authentication token!",
        },
        EDAMErrorCode.AUTH_EXPIRED: {
            "authenticationToken": "Authentication token expired or revoked!",
            "password": (
                "Password login disabled. Password reset required!\n"
                "Most probably, you log in to Evernote with Google or Apple account."
                " Use --oauth option to log in."
            ),
        },
    }

    if isinstance(exception, EDAMSystemException):
        error_param = exception.message
    else:
        error_param = exception.parameter

    try:
        error = messages[exception.errorCode][error_param]
    except KeyError:
        return

    raise EvernoteAuthError(error)



import logging
import time

from evernote.edam.type.ttypes import Note, Notebook

from evernote_backup.cli_app_util import is_output_to_terminal


def init_logging() -> None:
    for h in logging.root.handlers[:]:
        logging.root.removeHandler(h)
        h.close()


def init_logging_format() -> None:
    if is_output_to_terminal():
        log_format = "%(message)s"
    else:
        log_format = "%(asctime)s | [%(levelname)s] | %(thread)d | %(message)s"

    logging.basicConfig(format=log_format)


def log_format_note(note: Note) -> str:  # pragma: no cover
    n_info = [
        f"'{note.title}' [{note.guid}]",
        f"notebook_id [{note.notebookGuid}]",
    ]

    if not note.active:
        n_info.append("DELETED")

    return ", ".join(n_info)


def log_format_notebook(notebook: Notebook) -> str:  # pragma: no cover
    nb_info = [f"'{notebook.name}' [{notebook.guid}]"]

    if notebook.stack:
        nb_info.append(f"stack '{notebook.stack}'")

    return ", ".join(nb_info)


def get_time_txt(seconds: int) -> str:
    seconds_hour = 3600
    seconds_minute = 60

    if seconds > seconds_hour:
        return time.strftime("%H:%M:%S", time.gmtime(seconds))
    elif seconds > seconds_minute:
        return time.strftime("%M:%S", time.gmtime(seconds))

    return time.strftime("0:%S", time.gmtime(seconds))



import logging
from typing import Dict, Iterator, Optional

from evernote.edam.error.ttypes import EDAMNotFoundException
from evernote.edam.notestore import NoteStore
from evernote.edam.notestore.ttypes import SyncChunk
from evernote.edam.type.ttypes import LinkedNotebook, Note

from evernote_backup.evernote_client import EvernoteClient
from evernote_backup.evernote_client_util import NotebookAuth

logger = logging.getLogger(__name__)


class EvernoteClientSync(EvernoteClient):  # noqa: WPS214
    def __init__(
        self,
        backend: str,
        token: str,
        network_error_retry_count: int,
        max_chunk_results: int,
    ) -> None:
        super().__init__(
            backend=backend,
            token=token,
            network_error_retry_count=network_error_retry_count,
        )

        self._tags: Optional[dict] = None
        self._notebook_tags: Dict[str, Dict[str, str]] = {}
        self._linked_notebooks: Optional[dict] = None
        self.max_chunk_results = max_chunk_results

        self.shared_mode = False

    def get_note(self, note_guid: str) -> Note:
        logger.debug(f"Downloading note [{note_guid}]")

        note = self.note_store.getNote(
            note_guid, True, True, True, True  # noqa: WPS425
        )

        if note.tagGuids:
            if self.shared_mode:
                nb_tags = self.list_notebook_tags(note.notebookGuid)
                note.tagNames = [nb_tags[t] for t in note.tagGuids]
            else:
                note.tagNames = [self.tags[t] for t in note.tagGuids]

        logger.debug(f"Finished downloading note [{note.guid}]")

        return note

    def iter_sync_chunks(self, after_usn: int) -> Iterator[SyncChunk]:
        sync_filter = NoteStore.SyncChunkFilter(
            includeNotes=True,
            includeNoteResources=True,
            includeNoteAttributes=True,
            includeNotebooks=True,
            includeExpunged=True,
            includeLinkedNotebooks=True,
        )

        while True:
            chunk = self.note_store.getFilteredSyncChunk(
                after_usn, self.max_chunk_results, sync_filter
            )

            yield chunk

            after_usn = chunk.chunkHighUSN
            if chunk.chunkHighUSN == chunk.updateCount:
                return

    def iter_linked_notebook_sync_chunks(
        self, l_notebook: LinkedNotebook, after_usn: int
    ) -> Iterator[SyncChunk]:
        ln_note_store = self.get_note_store(l_notebook.shardId)
        is_full_sync = False

        while True:
            try:
                chunk = ln_note_store.getLinkedNotebookSyncChunk(
                    l_notebook, after_usn, self.max_chunk_results, is_full_sync
                )
            except EDAMNotFoundException:
                # Happens when linked notebook was unshared
                # just skip it, since expunging removed notebook will alter account data
                logger.warning(
                    f"Linked notebook '{l_notebook.shareName}' [{l_notebook.guid}]"
                    f" is not accessible, skipping..."
                )
                return

            if after_usn == chunk.updateCount:
                return

            yield chunk

            after_usn = chunk.chunkHighUSN
            if chunk.chunkHighUSN == chunk.updateCount:
                return

    def auth_linked_notebook(
        self, l_notebook_guid: str, notebook_guid: str
    ) -> NotebookAuth:
        l_notebook = self.linked_notebooks[l_notebook_guid]
        is_notebook_public = l_notebook.shareKey is None and l_notebook.uri is not None

        if logger.getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
            ln_info = f"{l_notebook.shareName} [{l_notebook.guid}]"
            if is_notebook_public:
                ln_info += " [PUBLIC]"  # noqa: WPS336

            logger.debug(f"Requesting access to linked notebook {ln_info}")

        ln_note_store = self.get_note_store(l_notebook.shardId)

        if is_notebook_public:
            auth_token = self.token
        else:
            auth_token = ln_note_store.authenticateToSharedNotebook(
                notebook_guid
            ).authenticationToken

        return NotebookAuth(token=auth_token, shard=l_notebook.shardId)

    @property
    def linked_notebooks(self) -> Dict[str, LinkedNotebook]:
        if self._linked_notebooks is None:
            self._linked_notebooks = {
                ln.guid: ln for ln in self.note_store.listLinkedNotebooks()
            }
        return self._linked_notebooks

    @property
    def tags(self) -> Dict[str, str]:
        if self._tags is None:
            self._tags = {t.guid: t.name for t in self.note_store.listTags()}
        return self._tags

    def list_notebook_tags(self, notebook_guid: str) -> Dict[str, str]:
        if notebook_guid not in self._notebook_tags:
            self._notebook_tags[notebook_guid] = {
                t.guid: t.name
                for t in self.note_store.listTagsByNotebook(notebook_guid)
            }
        return self._notebook_tags[notebook_guid]

    def get_remote_usn(self) -> int:
        return int(self.note_store.getSyncState().updateCount)



from typing import Optional, Tuple

import click
from evernote.edam.userstore.ttypes import AuthenticationResult

from evernote_backup.cli_app_auth_oauth import prompt_ota
from evernote_backup.cli_app_util import (
    ProgramTerminatedError,
    is_output_to_terminal,
    unscramble,
)
from evernote_backup.config import API_DATA
from evernote_backup.evernote_client_auth import EvernoteClientAuth
from evernote_backup.evernote_client_util import EvernoteAuthError


def get_auth_client(backend: str, network_retry_count: int) -> EvernoteClientAuth:
    key, secret = unscramble(API_DATA)

    return EvernoteClientAuth(
        consumer_key=key,
        consumer_secret=secret,
        backend=backend,
        network_error_retry_count=network_retry_count,
    )


def prompt_credentials(
    user: Optional[str],
    password: Optional[str],
) -> Tuple[str, str]:
    if not is_output_to_terminal() and not all([user, password]):
        raise ProgramTerminatedError("--user and --password are required!")

    if not user:
        user = str(click.prompt("Username or Email"))
    if not password:
        password = str(click.prompt("Password", hide_input=True))

    return user, password


def evernote_login_password(
    auth_user: Optional[str],
    auth_password: Optional[str],
    backend: str,
    network_retry_count: int,
) -> str:
    auth_user, auth_password = prompt_credentials(auth_user, auth_password)

    auth_client = get_auth_client(
        backend=backend, network_retry_count=network_retry_count
    )

    try:
        auth_res = auth_client.login(auth_user, auth_password)
    except EvernoteAuthError as e:
        raise ProgramTerminatedError(e)

    if auth_res.secondFactorRequired:
        auth_res = handle_two_factor_auth(
            auth_client,
            auth_res.authenticationToken,
            auth_res.secondFactorDeliveryHint,
        )

    return str(auth_res.authenticationToken)


def handle_two_factor_auth(
    auth_client: EvernoteClientAuth, token: str, delivery_hint: str
) -> AuthenticationResult:
    ota_code = prompt_ota(delivery_hint)

    try:
        return auth_client.two_factor_auth(token, ota_code)
    except EvernoteAuthError as e:
        raise ProgramTerminatedError(e)



from typing import Optional

from evernote_backup.cli_app_auth_oauth import evernote_login_oauth
from evernote_backup.cli_app_auth_password import evernote_login_password
from evernote_backup.cli_app_util import ProgramTerminatedError, logger
from evernote_backup.evernote_client_sync import EvernoteClientSync
from evernote_backup.evernote_client_util import EvernoteAuthError
from evernote_backup.token_util import get_token_expiration_date


def get_sync_client(
    auth_token: str,
    backend: str,
    network_error_retry_count: int,
    max_chunk_results: int,
) -> EvernoteClientSync:
    logger.info(f"Authorizing auth token, {backend} backend...")

    client = EvernoteClientSync(
        token=auth_token,
        backend=backend,
        network_error_retry_count=network_error_retry_count,
        max_chunk_results=max_chunk_results,
    )

    try:
        client.verify_token()
    except EvernoteAuthError as e:
        raise ProgramTerminatedError(e)

    token_expiration = get_token_expiration_date(auth_token)

    logger.info(f"Successfully authenticated as {client.user}!")
    logger.info(f"Current login will expire at {token_expiration}.")

    return client


def get_auth_token(
    auth_user: Optional[str],
    auth_password: Optional[str],
    auth_is_oauth: bool,
    auth_oauth_port: int,
    backend: str,
    network_retry_count: int,
) -> str:
    logger.info("Logging in to Evernote...")

    if auth_is_oauth:
        return evernote_login_oauth(backend, auth_oauth_port)

    return evernote_login_password(
        auth_user,
        auth_password,
        backend,
        network_retry_count,
    )



from evernote.edam.error.ttypes import EDAMUserException
from evernote.edam.userstore.ttypes import AuthenticationResult

from evernote_backup.evernote_client import EvernoteClient
from evernote_backup.evernote_client_util import raise_auth_error


class EvernoteClientAuth(EvernoteClient):
    def __init__(
        self,
        backend: str,
        network_error_retry_count: int,
        consumer_key: str,
        consumer_secret: str,
    ):
        super().__init__(  # noqa: S106
            backend=backend,
            token="",
            network_error_retry_count=network_error_retry_count,
        )

        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret

    def login(self, username: str, password: str) -> AuthenticationResult:
        try:
            return self.user_store.authenticateLongSessionV2(
                username=username,
                password=password,
                ssoLoginToken=None,
                consumerKey=self.consumer_key,
                consumerSecret=self.consumer_secret,
                deviceIdentifier=None,
                deviceDescription=self.device_description,
                supportsTwoFactor=True,
                supportsBusinessOnlyAccounts=True,
            )
        except EDAMUserException as e:
            raise_auth_error(e)
            raise

    def two_factor_auth(self, auth_token: str, ota_code: str) -> AuthenticationResult:
        try:
            return self.user_store.completeTwoFactorAuthentication(
                authenticationToken=auth_token,
                oneTimeCode=ota_code,
                deviceIdentifier=None,
                deviceDescription=self.device_description,
            )
        except EDAMUserException as e:
            raise_auth_error(e)
            raise



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['evernote_backup']

package_data = \
{'': ['*']}

install_requires = \
['click-option-group>=0.5.3,<0.6.0',
 'click>=8.0.1,<9.0.0',
 'evernote3>=1.25.14,<2.0.0',
 'oauth2>=1.9.0,<2.0.0',
 'xmltodict>=0.12.0,<0.13.0']

entry_points = \
{'console_scripts': ['evernote-backup = evernote_backup.cli:main']}

setup_kwargs = {
    'name': 'evernote-backup',
    'version': '1.7.1',
    'description': 'Backup & export all Evernote notes and notebooks.',
    'long_description': '# evernote-backup\n\n[![PyPI version](https://img.shields.io/pypi/v/evernote-backup?label=version)](https://pypi.python.org/pypi/evernote-backup)\n[![Python Version](https://img.shields.io/pypi/pyversions/evernote-backup.svg)](https://pypi.org/project/evernote-backup/)\n[![tests](https://github.com/vzhd1701/evernote-backup/actions/workflows/test.yml/badge.svg)](https://github.com/vzhd1701/evernote-backup/actions/workflows/test.yml)\n[![codecov](https://codecov.io/gh/vzhd1701/evernote-backup/branch/master/graph/badge.svg)](https://codecov.io/gh/vzhd1701/evernote-backup)\n\nBackup your notes & notebooks from Evernote locally and export them at any time!\n\n## Features\n\n- Quickly sync all your notes into the SQLite database for backup.\n- Export all backed up notes in `*.enex` format, as **notebooks** or **single notes**.\n- Support for both [Evernote](https://evernote.com/) and [Yinxiang (印象笔记)](https://yinxiang.com/).\n\n## Installation\n\n[**Download the latest release**](https://github.com/vzhd1701/evernote-backup/releases/latest) for your OS.\n\n### With PIP\n\n```bash\n$ pip install evernote-backup\n```\n\nOr, since **evernote-backup** is a standalone tool, it might be more convenient to install it using [**pipx**](https://github.com/pipxproject/pipx):\n\n```bash\n$ pipx install evernote-backup\n```\n\n### With Docker\n\n```bash\n$ docker run --rm -t -v "$PWD":/tmp ghcr.io/vzhd1701/evernote-backup:latest\n```\n\nTo log in to Evernote using OAuth with Docker, you\'ll have to forward port 10500 for a callback:\n\n```bash\n$ docker run --rm -t -v "$PWD":/tmp -p 10500:10500 ghcr.io/vzhd1701/evernote-backup:latest init-db --oauth\n```\n\n## Usage\n\n### Step 1. Database initialization\n\nTo start you need to initialize your database.\n\n```console\n$ evernote-backup init-db\nUsername or Email: user@example.com\nPassword:\nLogging in to Evernote...\nEnter one-time code: 120917\nAuthorizing auth token, evernote backend...\nSuccessfully authenticated as user!\nCurrent login will expire at 2022-03-10 10:22:00.\nInitializing database en_backup.db...\nReading database en_backup.db...\nSuccessfully initialized database for user!\n```\n\nBy default, it will prompt you to enter your account credentials. You can provide them beforehand with `--user` and `--password` options.\n\nIf you log in to Evernote with Google or Apple accounts, you must use the `--oauth` option.\n\nTo connect to **Yinxiang** instead of Evernote, use `--backend china` option. Unfortunately, OAuth is not supported for **Yinxiang** yet.\n\n### Step 2. Downloading Evernote data\n\nThen you will be able to sync your account data.\n\n```console\n$ evernote-backup sync\nReading database en_backup.db...\nAuthorizing auth token, evernote backend...\nSuccessfully authenticated as user!\nCurrent login will expire at 2022-03-10 10:22:00.\nSyncing latest changes...\n  [####################################]  6763/6763\n566 notes to download...\n  [####################################]  566/566\nUpdated or added notebooks: 23\nUpdated or added notes: 566\nExpunged notebooks: 0\nExpunged notes: 0\nSynchronization completed!\n```\n\nYou can interrupt this process at any point. It will continue from where it\'s left off when you will rerun `evernote-backup sync`.\n\n**evernote-backup** keeps track of the sync state and downloads only new changes that have been made since the last run. So every sync will go pretty fast, but you\'ll have to wait for a bit on the first run if you have a lot of notes in your account.\n\n### Step 3. Exporting `*.enex` files\n\nFinally, you can export your data into specified **output directory**\n\n```console\n$ evernote-backup export output_dir/\nReading database en_backup.db...\nExporting notes...\n  [####################################]  23/23\nAll notes have been exported!\n```\n\nBy default, **evernote-backup** will export notes by packing them into notebooks, one `*.enex` file each. If you want to extract notes as **separate files**, use the `--single-notes` flag.\n\nTo also include **trashed** notes in export, use the `--include-trash` flag.\n\nThat\'s it! So to export all your Evernote data, you will have to run three commands:\n\n```console\n$ evernote-backup init-db\n$ evernote-backup sync\n$ evernote-backup export output_dir/\n```\n\nAfter first initialization, you can schedule `evernote-backup sync` command to keep your local database always up-to-date.\n\n### How to refresh expired token\n\nIn case your auth token that you initialized your database with expires, you have an option to re-authorize it by running the `evernote-backup reauth` command. It has the same options as the `init-db` command.\n\n## Dependencies\n\n- [evernote3](https://github.com/evernote/evernote-sdk-python3) - to access Evernote API\n- [oauth2](https://github.com/joestump/python-oauth2) - to perform OAuth authentication\n- [xmltodict](https://github.com/martinblech/xmltodict) - to convert Evernote internal representation of notes into XML\n- [click](https://github.com/pallets/click) - to create a CLI interface\n- [click-option-group](https://github.com/click-contrib/click-option-group) - to add mutually exclusive options in сlick\n\n## Alternative tools\n\n**evernote-backup** is basically a clone of Evernote\'s original **ENScript** but simplified and stripped of its other functions. If you prefer to export your notes using Evernote\'s original tool, then you will need to take hold of [Evernote\'s legacy client](https://help.evernote.com/hc/en-us/articles/360052560314-Install-an-older-version-of-Evernote) and run the following commands:\n\n```console\nENScript.exe syncDatabase /d backup.ebx /u your@email.com /p your_password\nENScript.exe exportDatabase /d backup.ebx /f output_dir\n```\n\n### Further reading\n\n- [How to export Notebooks in new Evernote client](https://help.evernote.com/hc/en-us/articles/360053159414-Export-notebooks)\n- [Backing up and restoring Evernote data (Reference article) (requires registration)](https://discussion.evernote.com/forums/topic/86152-backing-up-and-restoring-evernote-data-reference-article/?tab=comments#comment-367110)\n- [Migrating Your Notes from Evernote to Obsidian](https://www.dmuth.org/migrating-from-evernote-to-obisidian/)\n- https://github.com/davidedc/A-thousand-notes\n\n## Similar projects\n\nI\'ve also combined a [bigger list](https://github.com/vzhd1701/evernote-backup/blob/master/SIMILAR_PROJECTS.md) where I included all adjacent projects that I could find.\n\n### Export\n\nProject                                                                |  Description                                                                                                      |  Language    |  Stars  |  Last commit\n-----------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------|--------------|---------|-------------\n[ExportAllEverNote](https://github.com/dong-s/ExportAllEverNote)       |  Export notes in ENEX format driectly from your account                                                           |  Python      |  13     |  2020\n[evernote2](https://github.com/JackonYang/evernote2)                   |  Improved version of standard Evernote SDK. Also provides a tool for exporting notes driectly from your account  |  Python      |  1      |  2020\n[evernote-to-sqlite](https://github.com/dogsheep/evernote-to-sqlite)   |  Converts ENEX files into SQLite database                                                                         |  Python      |  20     |  2021\n[enote](https://github.com/tkjacobsen/enote)                           |  Utility that can backup Evernote notes and notebooks                                                             |  Python      |  12     |  2017\n[evernote-exporter](https://github.com/shawndaniel/evernote-exporter)  |  Export notes from old Evernote local database .exb format                                                        |  Python      |  28     |  2016\n\n### Export / Sync\n\nProject                                                                |  Description                                                                                                      |  Language    |  Stars  |  Last commit\n-----------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------|--------------|---------|-------------\n[SyncToGit](https://github.com/KostyaEsmukov/SyncToGit)                |  Syncs your notes with their resources to the git repository in HTML format                                       |  Python      |  3      |  2020\n[evermark](https://github.com/akuma/evermark)                          |  A command line tool for syncing markdown notes to Evernote                                                       |  JavaScript  |  72     |  2018\n[eversync](https://github.com/yejianye/eversync)                       |  Sync your local directories with evernote notebooks                                                              |  Python      |  44     |  2017\n[EverMark](https://github.com/liuwons/EverMark)                        |  A tool that can sync local markdown/text notes to Evernote                                                       |  Python      |  50     |  2016\n[LocalEvernote](https://github.com/lwabish/LocalEvernote)              |  Syncs local directory containing notes in Markdown format with Evernote                                          |  Python      |  0      |  2021\n',
    'author': 'vzhd1701',
    'author_email': 'vzhd1701@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vzhd1701/evernote-backup',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
